"""Wrapper functions around an LLM chain."""

from typing import Any, Dict, Optional, Tuple

from langchain import LLMChain, OpenAI

from gpt_index.prompts.base import Prompt


def openai_llm_predict(
    prompt: Prompt, llm_args_dict: Optional[Dict] = None, **prompt_args: Any
) -> Tuple[str, str]:
    """Predict using OpenAI LLM with a prompt string.

    Also return the formatted prompt.

    """
    llm_args_dict = llm_args_dict or {}
    llm = OpenAI(temperature=0, **llm_args_dict)
    llm_chain = LLMChain(prompt=prompt, llm=llm)

    formatted_prompt = prompt.format(**prompt_args)
    full_prompt_args = prompt.get_full_format_args(prompt_args)
    return llm_chain.predict(**full_prompt_args), formatted_prompt



"""List index.

A simple data structure where GPT Index iterates through document chunks
in sequence in order to answer a given query.

"""

import json
from typing import Any, List, Optional

from gpt_index.constants import MAX_CHUNK_OVERLAP, MAX_CHUNK_SIZE, NUM_OUTPUTS
from gpt_index.indices.base import DEFAULT_MODE, BaseGPTIndex, BaseGPTIndexQuery
from gpt_index.indices.data_structs import IndexList
from gpt_index.indices.list.query import GPTListIndexQuery
from gpt_index.indices.utils import get_chunk_size_given_prompt, truncate_text
from gpt_index.langchain_helpers.text_splitter import TokenTextSplitter
from gpt_index.prompts.base import Prompt
from gpt_index.prompts.default_prompts import DEFAULT_TEXT_QA_PROMPT
from gpt_index.schema import Document


class GPTListIndex(BaseGPTIndex[IndexList]):
    """GPT List Index."""

    def __init__(
        self,
        documents: Optional[List[Document]] = None,
        index_struct: Optional[IndexList] = None,
        text_qa_template: Prompt = DEFAULT_TEXT_QA_PROMPT,
    ) -> None:
        """Initialize params."""
        self.text_qa_template = text_qa_template
        # we need to figure out the max length of refine_template or text_qa_template
        # to find the minimum chunk size.
        empty_qa = self.text_qa_template.format(context_str="", query_str="")
        chunk_size = get_chunk_size_given_prompt(
            empty_qa, MAX_CHUNK_SIZE, 1, NUM_OUTPUTS
        )
        self.text_splitter = TokenTextSplitter(
            separator=" ",
            chunk_size=chunk_size,
            chunk_overlap=MAX_CHUNK_OVERLAP,
        )
        super().__init__(documents=documents, index_struct=index_struct)

    def build_index_from_documents(self, documents: List[Document]) -> IndexList:
        """Build the index from documents."""
        # do a simple concatenation
        text_data = "\n".join([d.text for d in documents])
        index_struct = IndexList()
        text_chunks = self.text_splitter.split_text(text_data)
        for _, text_chunk in enumerate(text_chunks):
            fmt_text_chunk = truncate_text(text_chunk, 50)
            print(f"> Adding chunk: {fmt_text_chunk}")
            index_struct.add_text(text_chunk)
        return index_struct

    def _mode_to_query(self, mode: str, **query_kwargs: Any) -> BaseGPTIndexQuery:
        if mode == DEFAULT_MODE:
            query_kwargs.update(
                {
                    "text_qa_template": self.text_qa_template,
                }
            )
            query = GPTListIndexQuery(self.index_struct, **query_kwargs)
        else:
            raise ValueError(f"Invalid query mode: {mode}.")
        return query

    def insert(self, document: Document, **insert_kwargs: Any) -> None:
        """Insert a document."""
        text_chunks = self.text_splitter.split_text(document.text)
        for _, text_chunk in enumerate(text_chunks):
            fmt_text_chunk = truncate_text(text_chunk, 50)
            print(f"> Adding chunk: {fmt_text_chunk}")
            self._index_struct.add_text(text_chunk)

    def delete(self, document: Document) -> None:
        """Delete a document."""
        raise NotImplementedError("Delete not implemented for list index.")

    @classmethod
    def load_from_disk(cls, save_path: str, **kwargs: Any) -> "GPTListIndex":
        """Load from disk."""
        with open(save_path, "r") as f:
            return cls(index_struct=IndexList.from_dict(json.load(f)), **kwargs)



"""Test list index."""

from typing import Any, List
from unittest.mock import patch

import pytest

from gpt_index.indices.list.base import GPTListIndex
from gpt_index.langchain_helpers.text_splitter import TokenTextSplitter
from gpt_index.schema import Document
from tests.mock_utils.mock_text_splitter import mock_token_splitter_newline


@pytest.fixture
def documents() -> List[Document]:
    """Get documents."""
    # NOTE: one document for now
    doc_text = (
        "Hello world.\n"
        "This is a test.\n"
        "This is another test.\n"
        "This is a test v2."
    )
    return [Document(doc_text)]


@patch.object(TokenTextSplitter, "split_text", side_effect=mock_token_splitter_newline)
def test_build_list(_mock_splitter: Any, documents: List[Document]) -> None:
    """Test build list."""
    list_index = GPTListIndex(documents)
    assert len(list_index.index_struct.nodes) == 4
    # check contents of nodes
    assert list_index.index_struct.nodes[0].text == "Hello world."
    assert list_index.index_struct.nodes[1].text == "This is a test."
    assert list_index.index_struct.nodes[2].text == "This is another test."
    assert list_index.index_struct.nodes[3].text == "This is a test v2."


@patch.object(TokenTextSplitter, "split_text", side_effect=mock_token_splitter_newline)
def test_list_insert(_mock_splitter: Any, documents: List[Document]) -> None:
    """Test insert to list."""
    list_index = GPTListIndex([])
    assert len(list_index.index_struct.nodes) == 0
    list_index.insert(documents[0])
    # check contents of nodes
    assert list_index.index_struct.nodes[0].text == "Hello world."
    assert list_index.index_struct.nodes[1].text == "This is a test."
    assert list_index.index_struct.nodes[2].text == "This is another test."
    assert list_index.index_struct.nodes[3].text == "This is a test v2."



"""File for core data structures."""

import random
import sys
from dataclasses import dataclass, field
from typing import Dict, List, Optional, Set

from dataclasses_json import DataClassJsonMixin


@dataclass
class IndexStruct(DataClassJsonMixin):
    """A base data struct for a GPT index."""


@dataclass
class Node(IndexStruct):
    """A generic node of data.

    Used in the GPT Tree Index and List Index.

    """

    text: str
    # ID field
    index: int
    # used for GPTTreeIndex
    child_indices: Set[int]


@dataclass
class IndexGraph(IndexStruct):
    """A graph representing the tree-structured index."""

    all_nodes: Dict[int, Node]
    root_nodes: Dict[int, Node]

    @property
    def size(self) -> int:
        """Get the size of the graph."""
        return len(self.all_nodes)

    def get_children(self, parent_node: Optional[Node]) -> Dict[int, Node]:
        """Get nodes given indices."""
        if parent_node is None:
            return self.root_nodes
        else:
            return {i: self.all_nodes[i] for i in parent_node.child_indices}

    def insert_under_parent(self, node: Node, parent_node: Optional[Node]) -> None:
        """Insert under parent node."""
        if node.index in self.all_nodes:
            raise ValueError(
                "Cannot insert a new node with the same index as an existing node."
            )
        if parent_node is None:
            self.root_nodes[node.index] = node
        else:
            parent_node.child_indices.add(node.index)

        self.all_nodes[node.index] = node


@dataclass
class KeywordTable(IndexStruct):
    """A table of keywords mapping keywords to text chunks."""

    table: Dict[str, Set[int]] = field(default_factory=dict)
    text_chunks: Dict[int, str] = field(default_factory=dict)

    def _get_index(self) -> int:
        """Get the next index for the text chunk."""
        # randomly generate until we get a unique index
        while True:
            idx = random.randint(0, sys.maxsize)
            if idx not in self.text_chunks:
                break
        return idx

    def add_text(self, keywords: List[str], text_chunk: str) -> int:
        """Add text to table."""
        cur_idx = self._get_index()
        for keyword in keywords:
            if keyword not in self.table:
                self.table[keyword] = set()
            self.table[keyword].add(cur_idx)
        self.text_chunks[cur_idx] = text_chunk
        return cur_idx

    def get_texts(self, keyword: str) -> List[str]:
        """Get texts given keyword."""
        if keyword not in self.table:
            raise ValueError("Keyword not found in table.")
        return [self.text_chunks[idx] for idx in self.table[keyword]]

    @property
    def keywords(self) -> Set[str]:
        """Get all keywords in the table."""
        return set(self.table.keys())

    @property
    def size(self) -> int:
        """Get the size of the table."""
        return len(self.table)


@dataclass
class IndexList(IndexStruct):
    """A list of documents."""

    nodes: List[Node] = field(default_factory=list)

    def add_text(self, text_chunk: str) -> int:
        """Add text to table, return current position in list."""
        # don't worry about child indices for now, nodes are all in order
        cur_node = Node(text=text_chunk, index=len(self.nodes), child_indices=set())
        self.nodes.append(cur_node)
        return cur_node.index



"""Base data structure classes."""
import json
from abc import abstractmethod
from typing import Any, Generic, List, Optional, TypeVar, cast

from gpt_index.indices.data_structs import IndexStruct
from gpt_index.schema import Document

IS = TypeVar("IS", bound=IndexStruct)

DEFAULT_MODE = "default"


class BaseGPTIndexQuery(Generic[IS]):
    """Base GPT Index Query.

    Helper class that is used to query an index. Can be called within `query`
    method of a BaseGPTIndex object, or instantiated independently.

    """

    def __init__(
        self,
        index_struct: IS,
    ) -> None:
        """Initialize with parameters."""
        if index_struct is None:
            raise ValueError("index_struct must be provided.")
        self._validate_index_struct(index_struct)
        self._index_struct = index_struct

    @property
    def index_struct(self) -> IS:
        """Get the index struct."""
        return self._index_struct

    def _validate_index_struct(self, index_struct: IS) -> None:
        """Validate the index struct."""
        pass

    @abstractmethod
    def query(self, query_str: str, verbose: bool = False) -> str:
        """Answer a query."""


class BaseGPTIndex(Generic[IS]):
    """Base GPT Index."""

    def __init__(
        self,
        documents: Optional[List[Document]] = None,
        index_struct: Optional[IS] = None,
    ) -> None:
        """Initialize with parameters."""
        if index_struct is None and documents is None:
            raise ValueError("One of documents or index_struct must be provided.")
        if index_struct is not None and documents is not None:
            raise ValueError("Only one of documents or index_struct can be provided.")

        # build index struct in the init function
        if index_struct is not None:
            self._index_struct = index_struct
        else:
            documents = cast(List[Document], documents)
            self._index_struct = self.build_index_from_documents(documents)

    @property
    def index_struct(self) -> IS:
        """Get the index struct."""
        return self._index_struct

    @abstractmethod
    def build_index_from_documents(self, documents: List[Document]) -> IS:
        """Build the index from documents."""

    @abstractmethod
    def insert(self, document: Document, **insert_kwargs: Any) -> None:
        """Insert a document."""

    @abstractmethod
    def delete(self, document: Document) -> None:
        """Delete a document."""

    @abstractmethod
    def _mode_to_query(self, mode: str, **query_kwargs: Any) -> BaseGPTIndexQuery:
        """Query mode to class."""

    def query(
        self,
        query_str: str,
        verbose: bool = False,
        mode: str = DEFAULT_MODE,
        **query_kwargs: Any
    ) -> str:
        """Answer a query."""
        query_obj = self._mode_to_query(mode, **query_kwargs)
        return query_obj.query(query_str, verbose=verbose)

    @classmethod
    @abstractmethod
    def load_from_disk(cls, save_path: str, **kwargs: Any) -> "BaseGPTIndex":
        """Load from disk."""

    def save_to_disk(self, save_path: str) -> None:
        """Safe to file."""
        with open(save_path, "w") as f:
            json.dump(self.index_struct.to_dict(), f)



"""Simple reader that ."""
from pathlib import Path
from typing import Any, List

from gpt_index.readers.base import BaseReader
from gpt_index.schema import Document


class SimpleDirectoryReader(BaseReader):
    """Simple directory reader.

    Concatenates all files into one document text.

    """

    def __init__(self, input_dir: str) -> None:
        """Initialize with parameters."""
        self.input_dir = Path(input_dir)
        input_files = list(self.input_dir.iterdir())
        for input_file in input_files:
            if not input_file.is_file():
                raise ValueError(f"Expected {input_file} to be a file.")
        self.input_files = input_files

    def load_data(self, **load_kwargs: Any) -> List[Document]:
        """Load data from the input directory."""
        data = ""
        for input_file in self.input_files:
            with open(input_file, "r") as f:
                data += f.read()
            data += "\n"
        return [Document(data)]



"""Default query for GPTListIndex."""
from typing import Optional

from gpt_index.indices.base import BaseGPTIndexQuery
from gpt_index.indices.data_structs import IndexList
from gpt_index.indices.response_utils import give_response, refine_response
from gpt_index.indices.utils import truncate_text
from gpt_index.prompts.base import Prompt
from gpt_index.prompts.default_prompts import (
    DEFAULT_REFINE_PROMPT,
    DEFAULT_TEXT_QA_PROMPT,
)


class GPTListIndexQuery(BaseGPTIndexQuery[IndexList]):
    """GPTListIndex query."""

    def __init__(
        self,
        index_struct: IndexList,
        text_qa_template: Prompt = DEFAULT_TEXT_QA_PROMPT,
        refine_template: Prompt = DEFAULT_REFINE_PROMPT,
        keyword: Optional[str] = None,
    ) -> None:
        """Initialize params."""
        super().__init__(index_struct=index_struct)
        self.text_qa_template = text_qa_template
        self.refine_template = refine_template
        self.keyword = keyword

    def query(self, query_str: str, verbose: bool = False) -> str:
        """Answer a query."""
        print(f"> Starting query: {query_str}")
        response = None
        for node in self.index_struct.nodes:
            if self.keyword is not None and self.keyword not in node.text:
                continue
            fmt_text_chunk = truncate_text(node.text, 50)
            if verbose:
                print(f"> Searching in chunk: {fmt_text_chunk}")

            if response is None:
                response = give_response(
                    query_str,
                    node.text,
                    text_qa_template=self.text_qa_template,
                    refine_template=self.refine_template,
                    verbose=verbose,
                )
            else:
                response = refine_response(
                    response,
                    query_str,
                    node.text,
                    refine_template=self.refine_template,
                    verbose=verbose,
                )
            if verbose:
                print(f"> Response: {response}")
        return response or ""



"""Base schema for data structures."""
from dataclasses import dataclass
from typing import Dict, Optional, Set

from dataclasses_json import DataClassJsonMixin


@dataclass
class Document:
    """Generic interface for document."""

    text: str
    extra_info: Optional[Dict] = None


@dataclass
class Node(DataClassJsonMixin):
    """A node in the GPT tree index."""

    text: str
    index: int
    child_indices: Set[int]


@dataclass
class IndexGraph(DataClassJsonMixin):
    """A graph representing the tree-structured index."""

    all_nodes: Dict[int, Node]
    root_nodes: Dict[int, Node]

    @property
    def size(self) -> int:
        """Get the size of the graph."""
        return len(self.all_nodes)



"""Test tree index."""

from typing import Any, Dict, List, Optional, Tuple
from unittest.mock import patch

import pytest

from gpt_index.indices.data_structs import IndexGraph, Node
from gpt_index.indices.tree.base import GPTTreeIndex
from gpt_index.langchain_helpers.text_splitter import TokenTextSplitter
from gpt_index.schema import Document
from tests.mock_utils.mock_predict import mock_openai_llm_predict
from tests.mock_utils.mock_prompts import (
    MOCK_INSERT_PROMPT,
    MOCK_QUERY_PROMPT,
    MOCK_REFINE_PROMPT,
    MOCK_SUMMARY_PROMPT,
    MOCK_TEXT_QA_PROMPT,
)
from tests.mock_utils.mock_text_splitter import mock_token_splitter_newline


@pytest.fixture
def struct_kwargs() -> Tuple[Dict, Dict]:
    """Index kwargs."""
    index_kwargs = {
        "summary_template": MOCK_SUMMARY_PROMPT,
        "insert_prompt": MOCK_INSERT_PROMPT,
        "num_children": 2,
    }
    query_kwargs = {
        "query_template": MOCK_QUERY_PROMPT,
        "text_qa_template": MOCK_TEXT_QA_PROMPT,
        "refine_template": MOCK_REFINE_PROMPT,
    }
    return index_kwargs, query_kwargs


@pytest.fixture
def documents() -> List[Document]:
    """Get documents."""
    # NOTE: one document for now
    doc_text = (
        "Hello world.\n"
        "This is a test.\n"
        "This is another test.\n"
        "This is a test v2."
    )
    return [Document(doc_text)]


def _get_left_or_right_node(
    index_graph: IndexGraph, node: Optional[Node], left: bool = True
) -> Node:
    """Get 'left' or 'right' node."""
    if node is None:
        indices = set(index_graph.root_nodes.keys())
    else:
        indices = node.child_indices

    index = min(indices) if left else max(indices)

    if index not in index_graph.all_nodes:
        print(f"Node {index} not in index_graph.all_nodes")
    return index_graph.all_nodes[index]


@patch.object(TokenTextSplitter, "split_text", side_effect=mock_token_splitter_newline)
@patch("gpt_index.indices.tree.base.openai_llm_predict", mock_openai_llm_predict)
def test_build_tree(
    _mock_predict: Any, documents: List[Document], struct_kwargs: Dict
) -> None:
    """Test build tree."""
    index_kwargs, _ = struct_kwargs
    tree = GPTTreeIndex(documents, **index_kwargs)
    assert len(tree.index_struct.all_nodes) == 6
    # check contents of nodes
    assert tree.index_struct.all_nodes[0].text == "Hello world."
    assert tree.index_struct.all_nodes[1].text == "This is a test."
    assert tree.index_struct.all_nodes[2].text == "This is another test."
    assert tree.index_struct.all_nodes[3].text == "This is a test v2."
    assert tree.index_struct.all_nodes[4].text == ("Hello world.\nThis is a test.")
    assert tree.index_struct.all_nodes[5].text == (
        "This is another test.\nThis is a test v2."
    )


@patch.object(TokenTextSplitter, "split_text", side_effect=mock_token_splitter_newline)
@patch("gpt_index.indices.tree.base.openai_llm_predict", mock_openai_llm_predict)
@patch("gpt_index.indices.tree.leaf_query.openai_llm_predict", mock_openai_llm_predict)
def test_query(
    _mock_predict: Any, documents: List[Document], struct_kwargs: Dict
) -> None:
    """Test query."""
    index_kwargs, query_kwargs = struct_kwargs
    tree = GPTTreeIndex(documents, **index_kwargs)

    # test default query
    query_str = "What is?"
    response = tree.query(query_str, mode="default", **query_kwargs)
    assert response == ("What is?\n" "Hello world.")


@patch.object(TokenTextSplitter, "split_text", side_effect=mock_token_splitter_newline)
@patch("gpt_index.indices.tree.base.openai_llm_predict", mock_openai_llm_predict)
@patch("gpt_index.indices.tree.inserter.openai_llm_predict", mock_openai_llm_predict)
def test_insert(
    _mock_predict: Any, documents: List[Document], struct_kwargs: Dict
) -> None:
    """Test insert."""
    index_kwargs, _ = struct_kwargs
    tree = GPTTreeIndex(documents, **index_kwargs)

    # test insert
    new_doc = Document("This is a new doc.")
    tree.insert(new_doc)
    # Before:
    # Left root node: "Hello world.\nThis is a test."
    # "Hello world.", "This is a test" are two children of the left root node
    # After:
    # "Hello world.\nThis is a test\n.\nThis is a new doc." is the left root node
    # "Hello world", "This is a test\n.This is a new doc." are the children
    # of the left root node.
    # "This is a test", "This is a new doc." are the children of
    # "This is a test\n.This is a new doc."
    left_root = _get_left_or_right_node(tree.index_struct, None)
    assert left_root.text == "Hello world.\nThis is a test.\nThis is a new doc."
    left_root2 = _get_left_or_right_node(tree.index_struct, left_root)
    right_root2 = _get_left_or_right_node(tree.index_struct, left_root, left=False)
    assert left_root2.text == "Hello world."
    assert right_root2.text == "This is a test.\nThis is a new doc."
    left_root3 = _get_left_or_right_node(tree.index_struct, right_root2)
    right_root3 = _get_left_or_right_node(tree.index_struct, right_root2, left=False)
    assert left_root3.text == "This is a test."
    assert right_root3.text == "This is a new doc."

    # test insert from empty
    tree = GPTTreeIndex([], **index_kwargs)
    new_doc = Document("This is a new doc.")
    tree.insert(new_doc)
    assert len(tree.index_struct.all_nodes) == 1
    assert tree.index_struct.all_nodes[0].text == "This is a new doc."
